@@ -85,6 +85,8 @@ The convention of rotation around ``z``, followed by rotation around
 
 import math
 
+from .externals.six.moves import reduce
+
 import numpy as np
 
 
@@ -337,7 +339,7 @@ def quat2euler(q):
     large.
     '''
     # delayed import to avoid cyclic dependencies
-    import nibabel.quaternions as nq
+    from . import quaternions as nq
     return mat2euler(nq.quat2mat(q))
 
 
@@ -371,7 +373,7 @@ def euler2angle_axis(z=0, y=0, x=0):
     True
     '''
     # delayed import to avoid cyclic dependencies
-    import nibabel.quaternions as nq
+    from . import quaternions as nq
     return nq.quat2angle_axis(euler2quat(z, y, x))
 
 
@@ -409,6 +411,6 @@ def angle_axis2euler(theta, vector, is_normalized=False):
     repetition is large.
     '''
     # delayed import to avoid cyclic dependencies
-    import nibabel.quaternions as nq
+    from . import quaternions as nq
     M = nq.angle_axis2mat(theta, vector, is_normalized)
     return mat2euler(M)
