@@ -1,10 +1,13 @@
-from __future__ import with_statement
+from __future__ import division, print_function, absolute_import
 
 import numpy as np
 import getpass
 import time
 
 
+from .. externals.six.moves import xrange
+
+
 def _fread3(fobj):
     """Read a 3-byte int from an open binary file object
 
@@ -36,7 +39,7 @@ def _fread3_many(fobj, n):
         An array of 3 byte int
     """
     b1, b2, b3 = np.fromfile(fobj, ">u1", 3 * n).reshape(-1,
-                                                    3).astype(np.int).T
+                                                         3).astype(np.int).T
     return (b1 << 16) + (b2 << 8) + b3
 
 
@@ -82,8 +85,8 @@ def read_geometry(filepath):
                     nface += 1
 
         elif magic == 16777214:  # Triangle file
-            create_stamp = fobj.readline()
-            _ = fobj.readline()
+            fobj.readline()  # create_stamp
+            fobj.readline()
             vnum = np.fromfile(fobj, ">i4", 1)[0]
             fnum = np.fromfile(fobj, ">i4", 1)[0]
             coords = np.fromfile(fobj, ">f4", vnum * 3).reshape(vnum, 3)
@@ -113,11 +116,11 @@ def write_geometry(filepath, coords, faces, create_stamp=None):
 
     if create_stamp is None:
         create_stamp = "created by %s on %s" % (getpass.getuser(),
-            time.ctime())
+                                                time.ctime())
 
     with open(filepath, 'wb') as fobj:
         magic_bytes.tofile(fobj)
-        fobj.write("%s\n\n" % create_stamp)
+        fobj.write(("%s\n\n" % create_stamp).encode('utf-8'))
 
         np.array([coords.shape[0], faces.shape[0]], dtype='>i4').tofile(fobj)
 
@@ -152,7 +155,7 @@ def read_morph_data(filepath):
             curv = np.fromfile(fobj, ">f4", vnum)
         else:
             vnum = magic
-            _ = _fread3(fobj)
+            _fread3(fobj)
             curv = np.fromfile(fobj, ">i2", vnum) / 100
     return curv
 
@@ -163,24 +166,28 @@ def read_annot(filepath, orig_ids=False):
     Parameters
     ----------
     filepath : str
-        Path to annotation file
+        Path to annotation file.
     orig_ids : bool
         Whether to return the vertex ids as stored in the annotation
-        file or the positional colortable ids
+        file or the positional colortable ids. With orig_ids=False
+        vertices with no id have an id set to -1.
 
     Returns
     -------
-    labels : n_vtx numpy array
-        Annotation id at each vertex
-    ctab : numpy array
-        RGBA + label id colortable array
-
+    labels : ndarray, shape (n_vertices,)
+        Annotation id at each vertex. If a vertex does not belong
+        to any label and orig_ids=False, its id will be set to -1.
+    ctab : ndarray, shape (n_labels, 5)
+        RGBA + label id colortable array.
+    names : list of str
+        The names of the labels. The length of the list is n_labels.
     """
     with open(filepath, "rb") as fobj:
         dt = ">i4"
         vnum = np.fromfile(fobj, dt, 1)[0]
         data = np.fromfile(fobj, dt, vnum * 2).reshape(vnum, 2)
         labels = data[:, 1]
+
         ctab_exists = np.fromfile(fobj, dt, 1)[0]
         if not ctab_exists:
             raise Exception('Color table not found in annotation file')
@@ -207,37 +214,108 @@ def read_annot(filepath, orig_ids=False):
             n_entries = np.fromfile(fobj, dt, 1)[0]
             ctab = np.zeros((n_entries, 5), np.int)
             length = np.fromfile(fobj, dt, 1)[0]
-            _ = np.fromfile(fobj, "|S%d" % length, 1)[0]  # Orig table path
+            np.fromfile(fobj, "|S%d" % length, 1)[0]  # Orig table path
             entries_to_read = np.fromfile(fobj, dt, 1)[0]
             names = list()
             for i in xrange(entries_to_read):
-                _ = np.fromfile(fobj, dt, 1)[0]  # Structure
+                np.fromfile(fobj, dt, 1)[0]  # Structure
                 name_length = np.fromfile(fobj, dt, 1)[0]
                 name = np.fromfile(fobj, "|S%d" % name_length, 1)[0]
                 names.append(name)
                 ctab[i, :4] = np.fromfile(fobj, dt, 4)
                 ctab[i, 4] = (ctab[i, 0] + ctab[i, 1] * (2 ** 8) +
-                                ctab[i, 2] * (2 ** 16))
+                              ctab[i, 2] * (2 ** 16))
         ctab[:, 3] = 255
     if not orig_ids:
         ord = np.argsort(ctab[:, -1])
-        labels = ord[np.searchsorted(ctab[ord, -1], labels)]
+        mask = labels != 0
+        labels[~mask] = -1
+        labels[mask] = ord[np.searchsorted(ctab[ord, -1], labels[mask])]
     return labels, ctab, names
 
 
-def read_label(filepath):
+def write_annot(filepath, labels, ctab, names):
+    """Write out a Freesurfer annotation file.
+
+    See:
+    http://ftp.nmr.mgh.harvard.edu/fswiki/LabelsClutsAnnotationFiles#Annotation
+
+    Parameters
+    ----------
+    filepath : str
+        Path to annotation file to be written
+    labels : ndarray, shape (n_vertices,)
+        Annotation id at each vertex.
+    ctab : ndarray, shape (n_labels, 5)
+        RGBA + label id colortable array.
+    names : list of str
+        The names of the labels. The length of the list is n_labels.
+    """
+    with open(filepath, "wb") as fobj:
+        dt = ">i4"
+        vnum = len(labels)
+
+        def write(num, dtype=dt):
+            np.array([num]).astype(dtype).tofile(fobj)
+
+        def write_string(s):
+            write(len(s))
+            write(s, dtype='|S%d' % len(s))
+
+        # vtxct
+        write(vnum)
+
+        # convert labels into coded CLUT values
+        clut_labels = ctab[:, -1][labels]
+        clut_labels[np.where(labels == -1)] = 0
+
+        # vno, label
+        data = np.vstack((np.array(range(vnum)).astype(dt),
+                          clut_labels.astype(dt))).T
+        data.byteswap().tofile(fobj)
+
+        # tag
+        write(1)
+
+        # ctabversion
+        write(-2)
+
+        # maxstruc
+        write(np.max(labels) + 1)
+
+        # File of LUT is unknown.
+        write_string('NOFILE')
+
+        # num_entries
+        write(ctab.shape[0])
+
+        for ind, (clu, name) in enumerate(zip(ctab, names)):
+            write(ind)
+            write_string(name)
+            for val in clu[:-1]:
+                write(val)
+
+
+def read_label(filepath, read_scalars=False):
     """Load in a Freesurfer .label file.
 
     Parameters
     ----------
     filepath : str
         Path to label file
+    read_scalars : bool
+        If true, read and return scalars associated with each vertex
 
     Returns
     -------
     label_array : numpy array
         Array with indices of vertices included in label
+    scalar_array : numpy array (floats)
+        If read_scalars is True, array of scalar data for each vertex
 
     """
     label_array = np.loadtxt(filepath, dtype=np.int, skiprows=2, usecols=[0])
+    if read_scalars:
+        scalar_array = np.loadtxt(filepath, skiprows=2, usecols=[-1])
+        return label_array, scalar_array
     return label_array
